#include "proxmox-backup-client.h"
#include "vma.h"

#include "qemu/osdep.h"
#include "qemu/module.h"
#include "sysemu/block-backend.h"
#include "sysemu/blockdev.h"
#include "block/blockjob.h"
#include "qapi/qapi-commands-block.h"
#include "qapi/qmp/qerror.h"
#include "qemu/cutils.h"

#include <proxmox-backup-qemu.h>

/* PVE backup state and related function */

/*
 * Note: A resume from a qemu_coroutine_yield can happen in a different thread,
 * so you may not use normal mutexes within coroutines:
 *
 * ---bad-example---
 * qemu_rec_mutex_lock(lock)
 * ...
 * qemu_coroutine_yield() // wait for something
 * // we are now inside a different thread
 * qemu_rec_mutex_unlock(lock) // Crash - wrong thread!!
 * ---end-bad-example--
 *
 * ==> Always use CoMutext inside coroutines.
 * ==> Never acquire/release AioContext withing coroutines (because that use QemuRecMutex)
 *
 */

const char *PBS_BITMAP_NAME = "pbs-incremental-dirty-bitmap";

static struct PVEBackupState {
    struct {
        // Everything accessed from qmp_backup_query command is protected using
        // this lock. Do NOT hold this lock for long times, as it is sometimes
        // acquired from coroutines, and thus any wait time may block the guest.
        QemuMutex lock;
        Error *error;
        time_t start_time;
        time_t end_time;
        char *backup_file;
        uuid_t uuid;
        char uuid_str[37];
        size_t total;
        size_t dirty;
        size_t transferred;
        size_t reused;
        size_t zero_bytes;
        GList *bitmap_list;
        bool finishing;
        bool starting;
    } stat;
    int64_t speed;
    BackupPerf perf;
    VmaWriter *vmaw;
    ProxmoxBackupHandle *pbs;
    GList *di_list;
    JobTxn *txn;
    CoMutex backup_mutex;
    CoMutex dump_callback_mutex;
} backup_state;

static void pvebackup_init(void)
{
    qemu_mutex_init(&backup_state.stat.lock);
    qemu_co_mutex_init(&backup_state.backup_mutex);
    qemu_co_mutex_init(&backup_state.dump_callback_mutex);
}

// initialize PVEBackupState at startup
opts_init(pvebackup_init);

typedef struct PVEBackupDevInfo {
    BlockDriverState *bs;
    size_t size;
    uint64_t block_size;
    uint8_t dev_id;
    int completed_ret; // INT_MAX if not completed
    char targetfile[PATH_MAX];
    BdrvDirtyBitmap *bitmap;
    BlockDriverState *target;
    BlockJob *job;
} PVEBackupDevInfo;

static void pvebackup_propagate_error(Error *err)
{
    qemu_mutex_lock(&backup_state.stat.lock);
    error_propagate(&backup_state.stat.error, err);
    qemu_mutex_unlock(&backup_state.stat.lock);
}

static bool pvebackup_error_or_canceled(void)
{
    qemu_mutex_lock(&backup_state.stat.lock);
    bool error_or_canceled = !!backup_state.stat.error;
    qemu_mutex_unlock(&backup_state.stat.lock);

    return error_or_canceled;
}

static void pvebackup_add_transfered_bytes(size_t transferred, size_t zero_bytes, size_t reused)
{
    qemu_mutex_lock(&backup_state.stat.lock);
    backup_state.stat.zero_bytes += zero_bytes;
    backup_state.stat.transferred += transferred;
    backup_state.stat.reused += reused;
    qemu_mutex_unlock(&backup_state.stat.lock);
}

// This may get called from multiple coroutines in multiple io-threads
// Note1: this may get called after job_cancel()
static int coroutine_fn
pvebackup_co_dump_pbs_cb(
    void *opaque,
    uint64_t start,
    uint64_t bytes,
    const void *pbuf)
{
    assert(qemu_in_coroutine());

    const uint64_t size = bytes;
    const unsigned char *buf = pbuf;
    PVEBackupDevInfo *di = opaque;

    assert(backup_state.pbs);
    assert(buf);

    Error *local_err = NULL;
    int pbs_res = -1;

    bool is_zero_block = size == di->block_size && buffer_is_zero(buf, size);

    qemu_co_mutex_lock(&backup_state.dump_callback_mutex);

    // avoid deadlock if job is cancelled
    if (pvebackup_error_or_canceled()) {
        qemu_co_mutex_unlock(&backup_state.dump_callback_mutex);
        return -1;
    }

    uint64_t transferred = 0;
    uint64_t reused = 0;
    while (transferred < size) {
        uint64_t left = size - transferred;
        uint64_t to_transfer = left < di->block_size ? left : di->block_size;

        pbs_res = proxmox_backup_co_write_data(backup_state.pbs, di->dev_id,
            is_zero_block ? NULL : buf + transferred, start + transferred,
            to_transfer, &local_err);
        transferred += to_transfer;

        if (pbs_res < 0) {
            pvebackup_propagate_error(local_err);
            qemu_co_mutex_unlock(&backup_state.dump_callback_mutex);
            return pbs_res;
        }

        reused += pbs_res == 0 ? to_transfer : 0;
    }

    qemu_co_mutex_unlock(&backup_state.dump_callback_mutex);
    pvebackup_add_transfered_bytes(size, is_zero_block ? size : 0, reused);

    return size;
}

// This may get called from multiple coroutines in multiple io-threads
static int coroutine_fn
pvebackup_co_dump_vma_cb(
    void *opaque,
    uint64_t start,
    uint64_t bytes,
    const void *pbuf)
{
    assert(qemu_in_coroutine());

    const uint64_t size = bytes;
    const unsigned char *buf = pbuf;
    PVEBackupDevInfo *di = opaque;

    int ret = -1;

    assert(backup_state.vmaw);
    assert(buf);

    uint64_t remaining = size;

    uint64_t cluster_num = start / VMA_CLUSTER_SIZE;
    if ((cluster_num * VMA_CLUSTER_SIZE) != start) {
        Error *local_err = NULL;
        error_setg(&local_err,
                   "got unaligned write inside backup dump "
                   "callback (sector %ld)", start);
        pvebackup_propagate_error(local_err);
        return -1; // not aligned to cluster size
    }

    while (remaining > 0) {
        qemu_co_mutex_lock(&backup_state.dump_callback_mutex);
        // avoid deadlock if job is cancelled
        if (pvebackup_error_or_canceled()) {
            qemu_co_mutex_unlock(&backup_state.dump_callback_mutex);
            return -1;
        }

        size_t zero_bytes = 0;
        ret = vma_writer_write(backup_state.vmaw, di->dev_id, cluster_num, buf, &zero_bytes);
        qemu_co_mutex_unlock(&backup_state.dump_callback_mutex);

        ++cluster_num;
        buf += VMA_CLUSTER_SIZE;
        if (ret < 0) {
            Error *local_err = NULL;
            vma_writer_error_propagate(backup_state.vmaw, &local_err);
            pvebackup_propagate_error(local_err);
            return ret;
        } else {
            if (remaining >= VMA_CLUSTER_SIZE) {
                assert(ret == VMA_CLUSTER_SIZE);
                pvebackup_add_transfered_bytes(VMA_CLUSTER_SIZE, zero_bytes, 0);
                remaining -= VMA_CLUSTER_SIZE;
            } else {
                assert(ret == remaining);
                pvebackup_add_transfered_bytes(remaining, zero_bytes, 0);
                remaining = 0;
            }
        }
    }

    return size;
}

// assumes the caller holds backup_mutex
static void coroutine_fn pvebackup_co_cleanup(void)
{
    assert(qemu_in_coroutine());

    qemu_mutex_lock(&backup_state.stat.lock);
    backup_state.stat.finishing = true;
    qemu_mutex_unlock(&backup_state.stat.lock);

    if (backup_state.vmaw) {
        Error *local_err = NULL;
        vma_writer_close(backup_state.vmaw, &local_err);

        if (local_err != NULL) {
            pvebackup_propagate_error(local_err);
         }

        backup_state.vmaw = NULL;
    }

    if (backup_state.pbs) {
        if (!pvebackup_error_or_canceled()) {
            Error *local_err = NULL;
            proxmox_backup_co_finish(backup_state.pbs, &local_err);
            if (local_err != NULL) {
                pvebackup_propagate_error(local_err);
            }
        }

        proxmox_backup_disconnect(backup_state.pbs);
        backup_state.pbs = NULL;
    }

    g_list_free(backup_state.di_list);
    backup_state.di_list = NULL;

    qemu_mutex_lock(&backup_state.stat.lock);
    backup_state.stat.end_time = time(NULL);
    backup_state.stat.finishing = false;
    qemu_mutex_unlock(&backup_state.stat.lock);
}

static void coroutine_fn pvebackup_co_complete_stream(void *opaque)
{
    PVEBackupDevInfo *di = opaque;
    int ret = di->completed_ret;

    qemu_mutex_lock(&backup_state.stat.lock);
    bool starting = backup_state.stat.starting;
    qemu_mutex_unlock(&backup_state.stat.lock);
    if (starting) {
        /* in 'starting' state, no tasks have been run yet, meaning we can (and
         * must) skip all cleanup, as we don't know what has and hasn't been
         * initialized yet. */
        return;
    }

    qemu_co_mutex_lock(&backup_state.backup_mutex);

    if (ret < 0) {
        Error *local_err = NULL;
        error_setg(&local_err, "job failed with err %d - %s", ret, strerror(-ret));
        pvebackup_propagate_error(local_err);
    }

    di->bs = NULL;

    assert(di->target == NULL);

    bool error_or_canceled = pvebackup_error_or_canceled();

    if (backup_state.vmaw) {
        vma_writer_close_stream(backup_state.vmaw, di->dev_id);
    }

    if (backup_state.pbs && !error_or_canceled) {
        Error *local_err = NULL;
        proxmox_backup_co_close_image(backup_state.pbs, di->dev_id, &local_err);
        if (local_err != NULL) {
            pvebackup_propagate_error(local_err);
        }
    }

    if (di->job) {
        WITH_JOB_LOCK_GUARD() {
            job_unref_locked(&di->job->job);
            di->job = NULL;
        }
    }

    // remove self from job list
    backup_state.di_list = g_list_remove(backup_state.di_list, di);

    g_free(di);

    /* call cleanup if we're the last job */
    if (!g_list_first(backup_state.di_list)) {
        pvebackup_co_cleanup();
    }

    qemu_co_mutex_unlock(&backup_state.backup_mutex);
}

static void pvebackup_complete_cb(void *opaque, int ret)
{
    PVEBackupDevInfo *di = opaque;
    di->completed_ret = ret;

    /*
     * Schedule stream cleanup in async coroutine. close_image and finish might
     * take a while, so we can't block on them here. This way it also doesn't
     * matter if we're already running in a coroutine or not.
     * Note: di is a pointer to an entry in the global backup_state struct, so
     * it stays valid.
     */
    Coroutine *co = qemu_coroutine_create(pvebackup_co_complete_stream, di);
    aio_co_enter(qemu_get_aio_context(), co);
}

/*
 * job_cancel(_sync) does not like to be called from coroutines, so defer to
 * main loop processing via a bottom half. Assumes that caller holds
 * backup_mutex.
 */
static void job_cancel_bh(void *opaque) {
    CoCtxData *data = (CoCtxData*)opaque;

    /*
     * Be careful to pick a valid job to cancel:
     * 1. job_cancel_sync() does not expect the job to be finalized already.
     * 2. job_exit() might run between scheduling and running job_cancel_bh()
     *    and pvebackup_co_complete_stream() might not have removed the job from
     *    the list yet (in fact, cannot, because it waits for the backup_mutex).
     * Requiring !job_is_completed() ensures that no finalized job is picked.
     */
    GList *bdi = g_list_first(backup_state.di_list);
    while (bdi) {
        if (bdi->data) {
            BlockJob *bj = ((PVEBackupDevInfo *)bdi->data)->job;
            if (bj) {
                Job *job = &bj->job;
                WITH_JOB_LOCK_GUARD() {
                    if (!job_is_completed_locked(job)) {
                        job_cancel_sync_locked(job, true);
                        /*
                         * It's enough to cancel one job in the transaction, the
                         * rest will follow automatically.
                         */
                        break;
                    }
                }
            }
        }
        bdi = g_list_next(bdi);
    }

    aio_co_enter(data->ctx, data->co);
}

void coroutine_fn qmp_backup_cancel(Error **errp)
{
    Error *cancel_err = NULL;
    error_setg(&cancel_err, "backup canceled");
    pvebackup_propagate_error(cancel_err);

    qemu_co_mutex_lock(&backup_state.backup_mutex);

    if (backup_state.vmaw) {
        /* make sure vma writer does not block anymore */
        vma_writer_set_error(backup_state.vmaw, "backup canceled");
    }

    if (backup_state.pbs) {
        proxmox_backup_abort(backup_state.pbs, "backup canceled");
    }

    CoCtxData data = {
        .ctx = qemu_get_current_aio_context(),
        .co = qemu_coroutine_self(),
    };
    aio_bh_schedule_oneshot(data.ctx, job_cancel_bh, &data);
    qemu_coroutine_yield();

    qemu_co_mutex_unlock(&backup_state.backup_mutex);
}

// assumes the caller holds backup_mutex
static int coroutine_fn pvebackup_co_add_config(
    const char *file,
    const char *name,
    BackupFormat format,
    const char *backup_dir,
    VmaWriter *vmaw,
    ProxmoxBackupHandle *pbs,
    Error **errp)
{
    int res = 0;

    char *cdata = NULL;
    gsize clen = 0;
    GError *err = NULL;
    if (!g_file_get_contents(file, &cdata, &clen, &err)) {
        error_setg(errp, "unable to read file '%s'", file);
        return 1;
    }

    char *basename = g_path_get_basename(file);
    if (name == NULL) name = basename;

    if (format == BACKUP_FORMAT_VMA) {
        if (vma_writer_add_config(vmaw, name, cdata, clen) != 0) {
            error_setg(errp, "unable to add %s config data to vma archive", file);
            goto err;
        }
    } else if (format == BACKUP_FORMAT_PBS) {
        if (proxmox_backup_co_add_config(pbs, name, (unsigned char *)cdata, clen, errp) < 0)
            goto err;
    } else if (format == BACKUP_FORMAT_DIR) {
        char config_path[PATH_MAX];
        snprintf(config_path, PATH_MAX, "%s/%s", backup_dir, name);
        if (!g_file_set_contents(config_path, cdata, clen, &err)) {
            error_setg(errp, "unable to write config file '%s'", config_path);
            goto err;
        }
    }

 out:
    g_free(basename);
    g_free(cdata);
    return res;

 err:
    res = -1;
    goto out;
}

/*
 * backup_job_create can *not* be run from a coroutine (and requires an
 * acquired AioContext), so this can't either.
 * The caller is responsible that backup_mutex is held nonetheless.
 */
static void create_backup_jobs_bh(void *opaque) {

    assert(!qemu_in_coroutine());

    CoCtxData *data = (CoCtxData*)opaque;
    Error **errp = (Error**)data->data;

    Error *local_err = NULL;

    /* create job transaction to synchronize bitmap commit and cancel all
     * jobs in case one errors */
    if (backup_state.txn) {
        job_txn_unref(backup_state.txn);
    }
    backup_state.txn = job_txn_new_seq();

    /* create and start all jobs (paused state) */
    GList *l =  backup_state.di_list;
    while (l) {
        PVEBackupDevInfo *di = (PVEBackupDevInfo *)l->data;
        l = g_list_next(l);

        assert(di->target != NULL);

        MirrorSyncMode sync_mode = MIRROR_SYNC_MODE_FULL;
        BitmapSyncMode bitmap_mode = BITMAP_SYNC_MODE_NEVER;
        if (di->bitmap) {
            sync_mode = MIRROR_SYNC_MODE_BITMAP;
            bitmap_mode = BITMAP_SYNC_MODE_ON_SUCCESS;
        }
        AioContext *aio_context = bdrv_get_aio_context(di->bs);
        aio_context_acquire(aio_context);

        BlockJob *job = backup_job_create(
            NULL, di->bs, di->target, backup_state.speed, sync_mode, di->bitmap,
            bitmap_mode, false, NULL, &backup_state.perf, BLOCKDEV_ON_ERROR_REPORT,
            BLOCKDEV_ON_ERROR_REPORT, JOB_DEFAULT, pvebackup_complete_cb, di, backup_state.txn,
            &local_err);

        aio_context_release(aio_context);

        di->job = job;
        if (job) {
            WITH_JOB_LOCK_GUARD() {
                job_ref_locked(&job->job);
            }
        }

        if (!job || local_err) {
            error_setg(errp, "backup_job_create failed: %s",
                       local_err ? error_get_pretty(local_err) : "null");
            break;
        }

        bdrv_unref(di->target);
        di->target = NULL;
    }

    if (*errp) {
        /*
         * It's enough to cancel one job in the transaction, the rest will
         * follow automatically.
         */
        bool canceled = false;
        l = backup_state.di_list;
        while (l) {
            PVEBackupDevInfo *di = (PVEBackupDevInfo *)l->data;
            l = g_list_next(l);

            if (di->target) {
                bdrv_unref(di->target);
                di->target = NULL;
            }

            if (di->job) {
                WITH_JOB_LOCK_GUARD() {
                    if (!canceled) {
                        job_cancel_sync_locked(&di->job->job, true);
                        canceled = true;
                    }
                    job_unref_locked(&di->job->job);
                    di->job = NULL;
                }
            }
        }
    }

    /* return */
    aio_co_enter(data->ctx, data->co);
}

UuidInfo coroutine_fn *qmp_backup(
    const char *backup_file,
    bool has_password, const char *password,
    bool has_keyfile, const char *keyfile,
    bool has_key_password, const char *key_password,
    bool has_master_keyfile, const char *master_keyfile,
    bool has_fingerprint, const char *fingerprint,
    bool has_backup_ns, const char *backup_ns,
    bool has_backup_id, const char *backup_id,
    bool has_backup_time, int64_t backup_time,
    bool has_use_dirty_bitmap, bool use_dirty_bitmap,
    bool has_compress, bool compress,
    bool has_encrypt, bool encrypt,
    bool has_format, BackupFormat format,
    bool has_config_file, const char *config_file,
    bool has_firewall_file, const char *firewall_file,
    bool has_devlist, const char *devlist,
    bool has_speed, int64_t speed,
    bool has_max_workers, int64_t max_workers,
    Error **errp)
{
    assert(qemu_in_coroutine());

    qemu_co_mutex_lock(&backup_state.backup_mutex);

    BlockBackend *blk;
    BlockDriverState *bs = NULL;
    const char *backup_dir = NULL;
    Error *local_err = NULL;
    uuid_t uuid;
    VmaWriter *vmaw = NULL;
    ProxmoxBackupHandle *pbs = NULL;
    gchar **devs = NULL;
    GList *di_list = NULL;
    GList *l;
    UuidInfo *uuid_info;

    const char *config_name = "qemu-server.conf";
    const char *firewall_name = "qemu-server.fw";

    if (backup_state.di_list) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR,
                  "previous backup not finished");
        qemu_co_mutex_unlock(&backup_state.backup_mutex);
        return NULL;
    }

    /* Todo: try to auto-detect format based on file name */
    format = has_format ? format : BACKUP_FORMAT_VMA;

    if (has_devlist) {
        devs = g_strsplit_set(devlist, ",;:", -1);

        gchar **d = devs;
        while (d && *d) {
            blk = blk_by_name(*d);
            if (blk) {
                bs = blk_bs(blk);
                if (!bdrv_is_inserted(bs)) {
                    error_setg(errp, QERR_DEVICE_HAS_NO_MEDIUM, *d);
                    goto err;
                }
                PVEBackupDevInfo *di = g_new0(PVEBackupDevInfo, 1);
                di->bs = bs;
                di_list = g_list_append(di_list, di);
            } else {
                error_set(errp, ERROR_CLASS_DEVICE_NOT_FOUND,
                          "Device '%s' not found", *d);
                goto err;
            }
            d++;
        }

    } else {
        BdrvNextIterator it;

        bs = NULL;
        for (bs = bdrv_first(&it); bs; bs = bdrv_next(&it)) {
            if (!bdrv_is_inserted(bs) || bdrv_is_read_only(bs)) {
                continue;
            }

            PVEBackupDevInfo *di = g_new0(PVEBackupDevInfo, 1);
            di->bs = bs;
            di_list = g_list_append(di_list, di);
        }
    }

    if (!di_list) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR, "empty device list");
        goto err;
    }

    size_t total = 0;

    l = di_list;
    while (l) {
        PVEBackupDevInfo *di = (PVEBackupDevInfo *)l->data;
        l = g_list_next(l);
        if (bdrv_op_is_blocked(di->bs, BLOCK_OP_TYPE_BACKUP_SOURCE, errp)) {
            goto err;
        }

        ssize_t size = bdrv_getlength(di->bs);
        if (size < 0) {
            error_setg_errno(errp, -size, "bdrv_getlength failed");
            goto err;
        }
        di->size = size;
        total += size;

        di->completed_ret = INT_MAX;
    }

    uuid_generate(uuid);

    qemu_mutex_lock(&backup_state.stat.lock);
    backup_state.stat.reused = 0;

    /* clear previous backup's bitmap_list */
    if (backup_state.stat.bitmap_list) {
        GList *bl = backup_state.stat.bitmap_list;
        while (bl) {
            g_free(((PBSBitmapInfo *)bl->data)->drive);
            g_free(bl->data);
            bl = g_list_next(bl);
        }
        g_list_free(backup_state.stat.bitmap_list);
        backup_state.stat.bitmap_list = NULL;
    }

    if (format == BACKUP_FORMAT_PBS) {
        if (!has_password) {
            error_set(errp, ERROR_CLASS_GENERIC_ERROR, "missing parameter 'password'");
            goto err_mutex;
        }
        if (!has_backup_id) {
            error_set(errp, ERROR_CLASS_GENERIC_ERROR, "missing parameter 'backup-id'");
            goto err_mutex;
        }
        if (!has_backup_time) {
            error_set(errp, ERROR_CLASS_GENERIC_ERROR, "missing parameter 'backup-time'");
            goto err_mutex;
        }

        int dump_cb_block_size = PROXMOX_BACKUP_DEFAULT_CHUNK_SIZE; // Hardcoded (4M)
        firewall_name = "fw.conf";

        char *pbs_err = NULL;
        pbs = proxmox_backup_new_ns(
            backup_file,
            has_backup_ns ? backup_ns : NULL,
            backup_id,
            backup_time,
            dump_cb_block_size,
            has_password ? password : NULL,
            has_keyfile ? keyfile : NULL,
            has_key_password ? key_password : NULL,
            has_master_keyfile ? master_keyfile : NULL,
            has_compress ? compress : true,
            has_encrypt ? encrypt : has_keyfile,
            has_fingerprint ? fingerprint : NULL,
             &pbs_err);

        if (!pbs) {
            error_set(errp, ERROR_CLASS_GENERIC_ERROR,
                      "proxmox_backup_new failed: %s", pbs_err);
            proxmox_backup_free_error(pbs_err);
            goto err_mutex;
        }

        int connect_result = proxmox_backup_co_connect(pbs, errp);
        if (connect_result < 0)
            goto err_mutex;

        /* register all devices */
        l = di_list;
        while (l) {
            PVEBackupDevInfo *di = (PVEBackupDevInfo *)l->data;
            l = g_list_next(l);

            di->block_size = dump_cb_block_size;

            const char *devname = bdrv_get_device_name(di->bs);
            PBSBitmapAction action = PBS_BITMAP_ACTION_NOT_USED;
            size_t dirty = di->size;

            BdrvDirtyBitmap *bitmap = bdrv_find_dirty_bitmap(di->bs, PBS_BITMAP_NAME);
            bool expect_only_dirty = false;

            if (has_use_dirty_bitmap && use_dirty_bitmap) {
                if (bitmap == NULL) {
                    bitmap = bdrv_create_dirty_bitmap(di->bs, dump_cb_block_size, PBS_BITMAP_NAME, errp);
                    if (!bitmap) {
                        goto err_mutex;
                    }
                    action = PBS_BITMAP_ACTION_NEW;
                } else {
                    expect_only_dirty = proxmox_backup_check_incremental(pbs, devname, di->size) != 0;
                }

                if (expect_only_dirty) {
                    /* track clean chunks as reused */
                    dirty = MIN(bdrv_get_dirty_count(bitmap), di->size);
                    backup_state.stat.reused += di->size - dirty;
                    action = PBS_BITMAP_ACTION_USED;
                } else {
                    /* mark entire bitmap as dirty to make full backup */
                    bdrv_set_dirty_bitmap(bitmap, 0, di->size);
                    if (action != PBS_BITMAP_ACTION_NEW) {
                        action = PBS_BITMAP_ACTION_INVALID;
                    }
                }
                di->bitmap = bitmap;
            } else {
                /* after a full backup the old dirty bitmap is invalid anyway */
                if (bitmap != NULL) {
                    bdrv_release_dirty_bitmap(bitmap);
                    action = PBS_BITMAP_ACTION_NOT_USED_REMOVED;
                }
            }

            int dev_id = proxmox_backup_co_register_image(pbs, devname, di->size, expect_only_dirty, errp);
            if (dev_id < 0) {
                goto err_mutex;
            }

            if (!(di->target = bdrv_backup_dump_create(dump_cb_block_size, di->size, pvebackup_co_dump_pbs_cb, di, errp))) {
                goto err_mutex;
            }

            di->dev_id = dev_id;

            PBSBitmapInfo *info = g_malloc(sizeof(*info));
            info->drive = g_strdup(devname);
            info->action = action;
            info->size = di->size;
            info->dirty = dirty;
            backup_state.stat.bitmap_list = g_list_append(backup_state.stat.bitmap_list, info);
        }
    } else if (format == BACKUP_FORMAT_VMA) {
        vmaw = vma_writer_create(backup_file, uuid, &local_err);
        if (!vmaw) {
            if (local_err) {
                error_propagate(errp, local_err);
            }
            goto err_mutex;
        }

        /* register all devices for vma writer */
        l = di_list;
        while (l) {
            PVEBackupDevInfo *di = (PVEBackupDevInfo *)l->data;
            l = g_list_next(l);

            if (!(di->target = bdrv_backup_dump_create(VMA_CLUSTER_SIZE, di->size, pvebackup_co_dump_vma_cb, di, errp))) {
                goto err_mutex;
            }

            const char *devname = bdrv_get_device_name(di->bs);
            di->dev_id = vma_writer_register_stream(vmaw, devname, di->size);
            if (di->dev_id <= 0) {
                error_set(errp, ERROR_CLASS_GENERIC_ERROR,
                          "register_stream failed");
                goto err_mutex;
            }
        }
    } else if (format == BACKUP_FORMAT_DIR) {
        if (mkdir(backup_file, 0640) != 0) {
            error_setg_errno(errp, errno, "can't create directory '%s'\n",
                             backup_file);
            goto err_mutex;
        }
        backup_dir = backup_file;

        l = di_list;
        while (l) {
            PVEBackupDevInfo *di = (PVEBackupDevInfo *)l->data;
            l = g_list_next(l);

            const char *devname = bdrv_get_device_name(di->bs);
            snprintf(di->targetfile, PATH_MAX, "%s/%s.raw", backup_dir, devname);

            int flags = BDRV_O_RDWR;
            bdrv_img_create(di->targetfile, "raw", NULL, NULL, NULL,
                            di->size, flags, false, &local_err);
            if (local_err) {
                error_propagate(errp, local_err);
                goto err_mutex;
            }

            di->target = bdrv_open(di->targetfile, NULL, NULL, flags, &local_err);
            if (!di->target) {
                error_propagate(errp, local_err);
                goto err_mutex;
            }
        }
    } else {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR, "unknown backup format");
        goto err_mutex;
    }


    /* add configuration file to archive */
    if (has_config_file) {
        if (pvebackup_co_add_config(config_file, config_name, format, backup_dir,
                                    vmaw, pbs, errp) != 0) {
            goto err_mutex;
        }
    }

    /* add firewall file to archive */
    if (has_firewall_file) {
        if (pvebackup_co_add_config(firewall_file, firewall_name, format, backup_dir,
                                    vmaw, pbs, errp) != 0) {
            goto err_mutex;
        }
    }
    /* initialize global backup_state now */
    /* note: 'reused' and 'bitmap_list' are initialized earlier */

    if (backup_state.stat.error) {
        error_free(backup_state.stat.error);
        backup_state.stat.error = NULL;
    }

    backup_state.stat.start_time = time(NULL);
    backup_state.stat.end_time = 0;

    if (backup_state.stat.backup_file) {
        g_free(backup_state.stat.backup_file);
    }
    backup_state.stat.backup_file = g_strdup(backup_file);

    uuid_copy(backup_state.stat.uuid, uuid);
    uuid_unparse_lower(uuid, backup_state.stat.uuid_str);
    char *uuid_str = g_strdup(backup_state.stat.uuid_str);

    backup_state.stat.total = total;
    backup_state.stat.dirty = total - backup_state.stat.reused;
    backup_state.stat.transferred = 0;
    backup_state.stat.zero_bytes = 0;
    backup_state.stat.finishing = false;
    backup_state.stat.starting = true;

    qemu_mutex_unlock(&backup_state.stat.lock);

    backup_state.speed = (has_speed && speed > 0) ? speed : 0;

    backup_state.perf = (BackupPerf){ .max_workers = 16 };
    if (has_max_workers) {
        backup_state.perf.max_workers = max_workers;
    }

    backup_state.vmaw = vmaw;
    backup_state.pbs = pbs;

    backup_state.di_list = di_list;

    uuid_info = g_malloc0(sizeof(*uuid_info));
    uuid_info->UUID = uuid_str;

    /* Run create_backup_jobs_bh outside of coroutine (in BH) but keep
    * backup_mutex locked. This is fine, a CoMutex can be held across yield
    * points, and we'll release it as soon as the BH reschedules us.
    */
    CoCtxData waker = {
        .co = qemu_coroutine_self(),
        .ctx = qemu_get_current_aio_context(),
        .data = &local_err,
    };
    aio_bh_schedule_oneshot(waker.ctx, create_backup_jobs_bh, &waker);
    qemu_coroutine_yield();

    if (local_err) {
        error_propagate(errp, local_err);
        goto err;
    }

    qemu_co_mutex_unlock(&backup_state.backup_mutex);

    qemu_mutex_lock(&backup_state.stat.lock);
    backup_state.stat.starting = false;
    qemu_mutex_unlock(&backup_state.stat.lock);

    /* start the first job in the transaction */
    job_txn_start_seq(backup_state.txn);

    return uuid_info;

err_mutex:
    qemu_mutex_unlock(&backup_state.stat.lock);

err:

    l = di_list;
    while (l) {
        PVEBackupDevInfo *di = (PVEBackupDevInfo *)l->data;
        l = g_list_next(l);

        if (di->target) {
            bdrv_unref(di->target);
        }

        if (di->targetfile[0]) {
            unlink(di->targetfile);
        }
        g_free(di);
    }
    g_list_free(di_list);
    backup_state.di_list = NULL;

    if (devs) {
        g_strfreev(devs);
    }

    if (vmaw) {
        Error *err = NULL;
        vma_writer_close(vmaw, &err);
        unlink(backup_file);
    }

    if (pbs) {
        proxmox_backup_disconnect(pbs);
        backup_state.pbs = NULL;
    }

    if (backup_dir) {
        rmdir(backup_dir);
    }

    qemu_co_mutex_unlock(&backup_state.backup_mutex);
    return NULL;
}

BackupStatus *qmp_query_backup(Error **errp)
{
    BackupStatus *info = g_malloc0(sizeof(*info));

    qemu_mutex_lock(&backup_state.stat.lock);

    if (!backup_state.stat.start_time) {
        /* not started, return {} */
        qemu_mutex_unlock(&backup_state.stat.lock);
        return info;
    }

    info->has_status = true;
    info->has_start_time = true;
    info->start_time = backup_state.stat.start_time;

    if (backup_state.stat.backup_file) {
        info->has_backup_file = true;
        info->backup_file = g_strdup(backup_state.stat.backup_file);
    }

    info->has_uuid = true;
    info->uuid = g_strdup(backup_state.stat.uuid_str);

    if (backup_state.stat.end_time) {
        if (backup_state.stat.error) {
            info->status = g_strdup("error");
            info->has_errmsg = true;
            info->errmsg = g_strdup(error_get_pretty(backup_state.stat.error));
        } else {
            info->status = g_strdup("done");
        }
        info->has_end_time = true;
        info->end_time = backup_state.stat.end_time;
    } else {
        info->status = g_strdup("active");
    }

    info->has_total = true;
    info->total = backup_state.stat.total;
    info->has_dirty = true;
    info->dirty = backup_state.stat.dirty;
    info->has_zero_bytes = true;
    info->zero_bytes = backup_state.stat.zero_bytes;
    info->has_transferred = true;
    info->transferred = backup_state.stat.transferred;
    info->has_reused = true;
    info->reused = backup_state.stat.reused;
    info->finishing = backup_state.stat.finishing;

    qemu_mutex_unlock(&backup_state.stat.lock);

    return info;
}

PBSBitmapInfoList *qmp_query_pbs_bitmap_info(Error **errp)
{
    PBSBitmapInfoList *head = NULL, **p_next = &head;

    qemu_mutex_lock(&backup_state.stat.lock);

    GList *l = backup_state.stat.bitmap_list;
    while (l) {
        PBSBitmapInfo *info = (PBSBitmapInfo *)l->data;
        l = g_list_next(l);

        /* clone bitmap info to avoid auto free after QMP marshalling */
        PBSBitmapInfo *info_ret = g_malloc0(sizeof(*info_ret));
        info_ret->drive = g_strdup(info->drive);
        info_ret->action = info->action;
        info_ret->size = info->size;
        info_ret->dirty = info->dirty;

        PBSBitmapInfoList *info_list = g_malloc0(sizeof(*info_list));
        info_list->value = info_ret;

        *p_next = info_list;
        p_next = &info_list->next;
    }

    qemu_mutex_unlock(&backup_state.stat.lock);

    return head;
}

ProxmoxSupportStatus *qmp_query_proxmox_support(Error **errp)
{
    ProxmoxSupportStatus *ret = g_malloc0(sizeof(*ret));
    ret->pbs_library_version = g_strdup(proxmox_backup_qemu_version());
    ret->pbs_dirty_bitmap = true;
    ret->pbs_dirty_bitmap_savevm = true;
    ret->pbs_dirty_bitmap_migration = true;
    ret->query_bitmap_info = true;
    ret->pbs_masterkey = true;
    ret->backup_max_workers = true;
    return ret;
}
